// De.dll v0.0.02 Copyright(C) 2002 thejam79, 2003 minamina
// Avisynth Plugin - DeDot (YUY2 and YV12 Only, DeDot 0.3 base

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "DeDot.h"

extern "C" __declspec(dllexport)
const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
	DeDot::PluginInit(env);
	return 0;
}

DeDot::DeDot(PClip _child, const int nLuma2d, const int nLumaT, const int nChromaT1, const int nChromaT2, IScriptEnvironment* env)
	: GenericVideoFilter(_child), m_nLuma2d(nLuma2d), m_nLumaT(nLumaT), m_nChromaT1(nChromaT1), m_nChromaT2(nChromaT2)
{
	if(vi.IsYUY2() || vi.IsYV12())
	{
	}
	else
	{
		env->ThrowError("%s : YUY2, YV12 only", GetName());
	}
}

PVideoFrame __stdcall DeDot::GetFrame(int n, IScriptEnvironment *env)
{
	PVideoFrame frameDest = NULL;

	if (vi.IsYUY2())
	{
		frameDest = DeDot_YUY2(n, env);
	}
	else
	{
		frameDest = DeDot_YV12(n, env);
	}

	return frameDest;
}

#define CHROMA_MASK_T1_ASM(mmm, pC, pP) \
	__asm mov eax, pC \
	__asm mov ebx, pP \
	__asm movq mmm, [eax + esi * 8] \
	__asm movq mm6, [ebx + esi * 8] \
	__asm movq mm5, mmm				\
	__asm psubusb mmm, mm6			\
	__asm psubusb mm6, mm5			\
	__asm pmaxub mmm, mm6			\
	__asm psubusb mmm, chromaT1		\
	__asm pcmpeqb mmm, mm7			

#define CHROMA_MASK_T2_ASM(mmm, pC, pP, mmmabs) \
	__asm mov eax, pC \
	__asm mov ebx, pP \
	__asm movq mmm, [eax + esi * 8] \
	__asm movq mm6, [ebx + esi * 8] \
	__asm movq mm5, mmm				\
	__asm psubusb mmm, mm6			\
	__asm psubusb mm6, mm5			\
	__asm pmaxub mmm, mm6			\
	__asm movq mmmabs, mmm			\
	__asm psubusb mmm, chromaT2		\
	__asm pcmpeqb mmm, mm7			\
	__asm pcmpeqb mmm, mm7			

#define LUMA_MASK_T_ASM(mmm, pC, pP) \
	__asm mov eax, pC \
	__asm mov ebx, pP \
	__asm movq mmm, [eax + esi * 8] \
	__asm movq mm6, [ebx + esi * 8] \
	__asm movq mm5, mmm				\
	__asm psubusb mmm, mm6			\
	__asm psubusb mm6, mm5			\
	__asm pmaxub mmm, mm6			\
	__asm psubusb mmm, lumaT		\
	__asm pcmpeqb mmm, mm7			

PVideoFrame DeDot::DeDot_YUY2(int n, IScriptEnvironment *env)
{
	const __int64 maskY = 0x00ff00ff00ff00ffi64;
	const __int64 maskUV = 0xff00ff00ff00ff00i64;

	int x, y;

	BYTE chromaT1[8];
	BYTE chromaT2[8];
	WORD luma2d[4];
	WORD lumaT[4];
	BYTE bufAvg[8];

	for (x = 0; x < 8; x++)
	{
		chromaT1[x] = (BYTE)m_nChromaT1;
		chromaT2[x] = (BYTE)m_nChromaT2;
	}
	for (x = 0; x < 4; x++)
	{
		luma2d[x] = (WORD)m_nLuma2d;
		lumaT[x] = (WORD)m_nLumaT;
	}

	PVideoFrame srcPP(child->GetFrame(clipFrame(n - 2), env));
	PVideoFrame srcP(child->GetFrame(clipFrame(n - 1), env));
	PVideoFrame srcC(child->GetFrame(clipFrame(n), env));
	PVideoFrame srcN(child->GetFrame(clipFrame(n + 1), env));
	PVideoFrame srcNN(child->GetFrame(clipFrame(n + 2), env));
	PVideoFrame dst(env->NewVideoFrame(vi));

	const BYTE* pPP;
	const BYTE* pP;
	const BYTE* pC;
	const BYTE* pN;
	const BYTE* pNN;
	BYTE* pD;

	const int nPitch = srcC->GetPitch();
	const int nPitchDst = dst->GetPitch();
	const int nPitch2 = nPitch << 1;
	const int nRowSize = srcC->GetRowSize();
	const int nHeight = srcC->GetHeight();
	const int nEndX = (nRowSize >> 3) - 0; // 4 hbg邽

	env->BitBlt(DYP(dst, 0), nPitchDst, SYP(srcC, 0), nPitch, nRowSize, 2);
	env->BitBlt(DYP(dst, nHeight - 2), nPitchDst, SYP(srcC, nHeight - 2), nPitch, nRowSize, 2);

	for (y = 2; y < nHeight - 2; y++)
	{
		pPP = SYP(srcPP, y);
		pP = SYP(srcP, y);
		pC = SYP(srcC, y);
		pN = SYP(srcN, y);
		pNN = SYP(srcNN, y);

		pD = DYP(dst, y);
		__asm
		{
			pxor mm7, mm7
			xor esi, esi
			mov edx, pD
loopX:
			// chromaT1
			CHROMA_MASK_T1_ASM(mm0, pP, pN)
			movq mm4, mm0
			CHROMA_MASK_T1_ASM(mm0, pC, pPP)
			pand mm4, mm0
			CHROMA_MASK_T1_ASM(mm0, pC, pNN)
			pand mm4, mm0

			// chromaT2
			CHROMA_MASK_T2_ASM(mm0, pC, pP, mm2)
			pand mm4, mm0
			CHROMA_MASK_T2_ASM(mm0, pC, pN, mm3)
			pand mm4, mm0

			// chroma
			mov eax, pC
			mov ebx, pP
			mov ecx, pN
			movq mm6, [eax + esi * 8]	;
			movq mm0, [ebx + esi * 8]	;
			movq mm1, [ecx + esi * 8]	;
			pavgb mm0, mm6				;
			pavgb mm1, mm6				;
			psubusb mm3, mm2			;
			pcmpeqb mm3, mm7			;
			pand mm1, mm3				;
			pandn mm3, mm0				;
			por mm1, mm3				;
			movq bufAvg, mm1			;
			pand mm1, mm4				;
			pandn mm4, mm6				;
			por mm4, mm1				;
			pand mm6, maskY				;
			pand mm4, maskUV			;
			por mm6, mm4				;
			movq mm3, mm6				;


			// luma2d
			movq mm0, [eax + esi * 8 - 2]	;
			movq mm1, mm6					;
			pand mm0, maskY					;
			pand mm1, maskY					;
			movq mm4, [eax + esi * 8 + 2]	;
			pand mm4, maskY					;
			paddw mm0, mm4					;
			mov ecx, nPitch2				;
			mov ebx, eax					;
			sub ebx, ecx					;
			movq mm2, [ebx + esi * 8]		;
			pand mm2, maskY					;
			lea ebx, [eax + ecx]			;
			movq mm4, [ebx + esi * 8]		;
			pand mm4, maskY					;

			paddw mm2, mm4					;
			paddw mm1, mm1					;
			psubw mm0, mm1					;
			psubw mm2, mm1					;
			pxor mm1, mm1					;
			pxor mm4, mm4					;
			psubw mm1, mm0					;
			psubw mm4, mm2					;
			pmaxsw mm0, mm1					;
			pmaxsw mm2, mm4					;
			psubusw mm0, luma2d				;
			psubusw mm2, luma2d				;
			por mm2, mm0					;
			pcmpeqw mm2, mm7				;
			pcmpeqw mm2, mm7				;
			movq mm4, mm2					;

			movd eax, mm2					;
			punpckhdq mm2, mm7				;
			movd ecx, mm2					;
			or eax, ecx						;
			jz output						;

			// lumaT
			LUMA_MASK_T_ASM(mm0, pP, pN)
			pand mm4, mm0
			LUMA_MASK_T_ASM(mm0, pC, pPP)
			pand mm4, mm0
			LUMA_MASK_T_ASM(mm0, pC, pNN)
			pand mm4, mm0

			// mov mm3, pCf					;
			movq mm6, bufAvg				;
			pand mm6, mm4					;
			pandn mm4, mm3					;
			por mm4, mm6					;
			pand mm4, maskY					;
			pand mm3, maskUV				;
			por mm3, mm4					;
output:
			movntq [edx + esi * 8], mm3	;

			inc esi						;
			cmp esi, nEndX
			jl loopX
		}
	}
	USE_MMX2;

	return dst;
}

#define CHROMA_LAST_ASM(mmm, pC, pP, pN) \
	__asm mov eax, pC \
	__asm mov ebx, pP \
	__asm mov edi, pN \
	__asm movq mm6, [eax + esi * 8]	\
	__asm movq mm0, [ebx + esi * 8]	\
	__asm movq mm1, [edi + esi * 8]	\
	__asm pavgb mm0, mm6			\
	__asm pavgb mm1, mm6			\
	__asm pcmpeqb mm2, mm7			\
	__asm pand mm1, mm2				\
	__asm pandn mm2, mm0			\
	__asm por mm1, mm2				\
	__asm pand mm1, mmm				\
	__asm pandn mmm, mm6			\
	__asm por mmm, mm1				

#define ABS_ASM(mmm, mmn) \
	__asm movq mm6, mmm				\
	__asm psubusb mmm, mmn			\
	__asm psubusb mmn, mm6			\
	__asm pmaxub mmm, mmn			

PVideoFrame DeDot::DeDot_YV12(int n, IScriptEnvironment *env)
{
	int x, y;

	BYTE chromaT1[8];
	BYTE chromaT2[8];
	WORD luma2d[4];
	BYTE lumaT[8];

	for (x = 0; x < 8; x++)
	{
		chromaT1[x] = (BYTE)m_nChromaT1;
		chromaT2[x] = (BYTE)m_nChromaT2;
		lumaT[x] = (BYTE)m_nLumaT;
	}
	for (x = 0; x < 4; x++)
	{
		luma2d[x] = (WORD)m_nLuma2d;
	}

	PVideoFrame srcPP(child->GetFrame(clipFrame(n - 2), env));
	PVideoFrame srcP(child->GetFrame(clipFrame(n - 1), env));
	PVideoFrame srcC(child->GetFrame(clipFrame(n), env));
	PVideoFrame srcN(child->GetFrame(clipFrame(n + 1), env));
	PVideoFrame srcNN(child->GetFrame(clipFrame(n + 2), env));
	PVideoFrame dst(env->NewVideoFrame(vi));

	const BYTE* pPP_U;
	const BYTE* pP_U;
	const BYTE* pC_U;
	const BYTE* pN_U;
	const BYTE* pNN_U;
	BYTE* pD_U;
	const BYTE* pPP_V;
	const BYTE* pP_V;
	const BYTE* pC_V;
	const BYTE* pN_V;
	const BYTE* pNN_V;
	BYTE* pD_V;

//	const int nPitch_U = srcC->GetPitch(PLANAR_U);
//	const int nPitch2 = nPitch << 1;
	const int nRowSize_U = srcC->GetRowSize(PLANAR_U);
	const int nHeight_U = srcC->GetHeight(PLANAR_U);
	const int nEndX_U = nRowSize_U >> 3; // 4 hbg邽

	// chroma loop
	for (y = 0; y < nHeight_U; y++)
	{
		pPP_U = SYP(srcPP, y, PLANAR_U);
		pP_U = SYP(srcP, y, PLANAR_U);
		pC_U = SYP(srcC, y, PLANAR_U);
		pN_U = SYP(srcN, y, PLANAR_U);
		pNN_U = SYP(srcNN, y, PLANAR_U);
		pD_U = DYP(dst, y, PLANAR_U);
		pPP_V = SYP(srcPP, y, PLANAR_V);
		pP_V = SYP(srcP, y, PLANAR_V);
		pC_V = SYP(srcC, y, PLANAR_V);
		pN_V = SYP(srcN, y, PLANAR_V);
		pNN_V = SYP(srcNN, y, PLANAR_V);
		pD_V = DYP(dst, y, PLANAR_V);

		__asm
		{
			pxor mm7, mm7
			mov ecx, pD_U				;
			mov edx, pD_V				;
			xor esi, esi				;
loopXC:
			// chromaT1
			CHROMA_MASK_T1_ASM(mm3, pP_U, pN_U)
			CHROMA_MASK_T1_ASM(mm4, pP_V, pN_V)
			CHROMA_MASK_T1_ASM(mm0, pC_U, pPP_U)
			pand mm3, mm0
			CHROMA_MASK_T1_ASM(mm0, pC_V, pPP_V)
			pand mm4, mm0
			CHROMA_MASK_T1_ASM(mm0, pC_U, pNN_U)
			pand mm3, mm0
			CHROMA_MASK_T1_ASM(mm0, pC_V, pNN_V)
			pand mm4, mm0

			// chromaT2 U
			CHROMA_MASK_T2_ASM(mm0, pC_U, pP_U, mm1)
			pand mm3, mm0
			CHROMA_MASK_T2_ASM(mm0, pC_U, pN_U, mm2)
			pand mm3, mm0
			psubusb mm2, mm1			;

			// chroma U
			CHROMA_LAST_ASM(mm3, pC_U, pP_U, pN_U) // mm2
			movntq [ecx + esi * 8], mm3	;

			// chromaT2 V
			CHROMA_MASK_T2_ASM(mm0, pC_V, pP_V, mm1)
			pand mm4, mm0
			CHROMA_MASK_T2_ASM(mm0, pC_V, pN_V, mm2)
			pand mm4, mm0
			psubusb mm2, mm1			;

			// chroma V
			CHROMA_LAST_ASM(mm4, pC_V, pP_V, pN_V) // mm2
			movntq [edx + esi * 8], mm4	;

			inc esi
			cmp esi, nEndX_U
			jl loopXC
		}
	}

	const BYTE* pPP;
	const BYTE* pP;
	const BYTE* pC;
	const BYTE* pN;
	const BYTE* pNN;
	BYTE* pD;

	const int nPitch = srcC->GetPitch();
	const int nPitchDst = dst->GetPitch();
	const int nPitch2 = nPitch << 1;
	const int nRowSize = srcC->GetRowSize();
	const int nHeight = srcC->GetHeight();
	const int nEndX = (nRowSize >> 3) - 0; // 4 hbg邽

	env->BitBlt(DYP(dst, 0), nPitchDst, SYP(srcC, 0), nPitch, nRowSize, 2);
	env->BitBlt(DYP(dst, nHeight - 2), nPitchDst, SYP(srcC, nHeight - 2), nPitch, nRowSize, 2);

	// luma loop
	for (y = 2; y < nHeight - 2; y++)
	{
		pPP = SYP(srcPP, y);
		pP = SYP(srcP, y);
		pC = SYP(srcC, y);
		pN = SYP(srcN, y);
		pNN = SYP(srcNN, y);
		pD = DYP(dst, y);

		__asm
		{
			pxor mm7, mm7
			mov ecx, pC					;
			mov edx, pD					;
			mov edi, nPitch2			;
			xor esi, esi				;
loopXL:
			// luma2d
			movq mm0, [ecx + esi * 8 - 1]	;
			movq mm5, mm0					;
			punpcklbw mm0, mm7				;
			punpckhbw mm5, mm7				;
			movq mm4, [ecx + esi * 8 + 1]	;
			movq mm6, mm4					;
			punpcklbw mm4, mm7				;
			punpckhbw mm6, mm7				;
			paddw mm0, mm4					;
			paddw mm5, mm6					;

			mov ebx, ecx					;
			sub ebx, edi					;
			movq mm2, [ebx + esi * 8]		;
			movq mm3, mm2					;
			punpcklbw mm2, mm7				;
			punpckhbw mm3, mm7				;
			lea ebx, [ecx + edi]			;
			movq mm4, [ebx + esi * 8]		;
			movq mm6, mm4					;
			punpcklbw mm4, mm7				;
			punpckhbw mm6, mm7				;
			paddw mm2, mm4					;
			paddw mm3, mm6					;

			movq mm1, [ecx + esi * 8]		;
			movq mm6, mm1					;
			punpcklbw mm1, mm7				;
			punpckhbw mm6, mm7				;
			paddw mm1, mm1					;
			paddw mm6, mm6					;
			psubw mm0, mm1					;
			psubw mm5, mm6					;
			psubw mm2, mm1					;
			psubw mm3, mm6					;

			pxor mm1, mm1					;
			pxor mm4, mm4					;
			psubw mm1, mm0					;
			psubw mm4, mm2					;
			pmaxsw mm0, mm1					;
			pmaxsw mm2, mm4					;
			psubusw mm0, luma2d				;
			psubusw mm2, luma2d				;
			por mm2, mm0					;

			pxor mm1, mm1					;
			pxor mm4, mm4					;
			psubw mm1, mm5					;
			psubw mm4, mm3					;
			pmaxsw mm5, mm1					;
			pmaxsw mm3, mm4					;
			psubusw mm5, luma2d				;
			psubusw mm3, luma2d				;
			por mm3, mm5					;

			packuswb mm2, mm3				;

			pcmpeqb mm2, mm7				;
			pcmpeqb mm2, mm7				;
			movq mm4, mm2					;

			movd eax, mm2					;
			punpckhdq mm2, mm7				;
			movd ebx, mm2					;
			movq mm3, [ecx + esi * 8]		;
			or eax, ebx						;
			jz output						;

			// lumaT
			LUMA_MASK_T_ASM(mm0, pP, pN)
			pand mm4, mm0
			LUMA_MASK_T_ASM(mm0, pC, pPP)
			pand mm4, mm0
			LUMA_MASK_T_ASM(mm0, pC, pNN)
			pand mm4, mm0

			// luma avg
			mov eax, pP
			movq mm0, [eax + esi * 8]		;
			movq mm5, mm0					;
			movq mm2, mm3					;
			ABS_ASM(mm5, mm2)				;
			mov ebx, pN
			movq mm1, [ebx + esi * 8]		;
			movq mm2, mm1					;
			pavgb mm0, mm3					;
			pavgb mm1, mm3					;
			ABS_ASM(mm2, mm3)				;
			psubusb mm2, mm5				;
			pcmpeqb mm2, mm7				;
			pand mm1, mm2					;
			pandn mm2, mm0					;
			por mm1, mm2					;

			movq mm3, [ecx + esi * 8]		;
			pand mm1, mm4					;
			pandn mm4, mm3					;
			por mm4, mm1					;
			movq mm3, mm4
output:
			movq [edx + esi * 8], mm3

			inc esi
			cmp esi, nEndX
			jl loopXL
		}
	}
	USE_MMX2;

	return dst;
}
